#ifndef __PASS_BAND_FILTER__
#define __PASS_BAND_FILTER__

/*!
  \file
\verbatim

Copyright (c) 2004, Sylvain Paris and Francois Sillion
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    * Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
    
    * Redistributions in binary form must reproduce the above
    copyright notice, this list of conditions and the following
    disclaimer in the documentation and/or other materials provided
    with the distribution.

    * Neither the name of ARTIS, GRAVIR-IMAG nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

\endverbatim
*  This file provides some pass-band filters. For now, only the difference
* of Gaussians is implemented. Several sets of parameters are provided for
* convinience.
*
*  <b>This file requires the Function_2D module that can be downloaded from
* <a href="http://artis.imag.fr/Members/Sylvain.Paris/index.html#code">Sylvain Paris' page</a>.
*/


#include <complex>

#include "function_2D.h"

#include "fftw3.h"

#include "fft/support.h"


namespace FFT{


  //! Filter f by (Gaussian_small - Gaussian_large).
  template<typename Function>
  inline void difference_of_gaussians(const Function& f,
				      const float     small_wave_length,
				      const float     large_wave_length,
				      Function* const g);



  //! Filter f by (Gaussian_small - Gaussian_large). Support is given,
  //! goes faster for multiple computations.
  template<typename Function>
  inline void  difference_of_gaussians(const Function& f,
				       const float     small_wave_length,
				       const float     large_wave_length,
				       Function* const g,
				       Support&        support);

  //! Filter f by a pass-band filter centered on \p center_wave_length.
  template<typename Function>
  inline void difference_of_gaussians(const Function& f,
				      const float     center_wave_length,
				      Function* const g);



  //! Filter f by a pass-band filter centered on \p
  //! center_wave_length. Support is given, goes faster for multiple
  //! computations.
  template<typename Function>
  inline void  difference_of_gaussians(const Function& f,
				       const float     center_wave_length,
				       Function* const g,
				       Support&        support);


  //! Filter f (represented by its Fourier transform) by a pass-band
  //! filter centered on \p center_wave_length.
  template<typename Function>
  inline void difference_of_gaussians(const Support::buffer_type& F,
				      const float     center_wave_length,
				      Function* const g,
				      Support&        support);

  
  
  //! This function mixes frequency representation for f and space
  //! representation for g to provide a fast process when f filtered several
  //! times.
  //! Filter f by (Gaussian_small - Gaussian_large).
  //! f is given through its Fourier transform F in a Support::buffer_type.
  template<typename Function>
  void difference_of_gaussians(const Support::buffer_type& F,
			       const float     small_wave_length,
			       const float     large_wave_length,
			       Function* const g,
			       Support&        support);




  //! Functions purely in Fourier domain.
  namespace Fourier_domain{ 
  
    //! Filter f (represented by its Fourier transform) by a pass-band
    //! filter centered on \p center_wave_length.
    //! Return value G in Fourier domain.
    inline void
    difference_of_gaussians(const Support::buffer_type& F,
			    const float                 center_wave_length,
			    Support::buffer_type* const G,
			    Support&                    support);
  
    //! This function works only in fourier domain.
    //! Filter f by (Gassian_small - Gaussian_large).
    //! f is given through its Fourier transform F in a Support::buffer_type.
    //! g is given back through its Fourier transform G in a Support::buffer_type.
    inline void
    difference_of_gaussians(const Support::buffer_type& F,
			    const float                 small_wave_length,
			    const float                 large_wave_length,
			    Support::buffer_type* const G,
			    Support&                    support);
  }


  

/*
  ###########################################################
  ###########################################################
  ###########################################################
  ##############                               ##############
  ##############  I M P L E M E N T A T I O N  ##############
  ##############                               ##############
  ###########################################################
  ###########################################################
  ###########################################################
*/

  


  template<typename Function>
  void difference_of_gaussians(const Function& f,
			       const float     center_wave_length,
			       Function* const g){

    difference_of_gaussians(f,
			    center_wave_length*M_SQRT1_2*M_SQRT1_2,
			    center_wave_length*M_SQRT2*M_SQRT1_2,
			    g);

  }



  template<typename Function>
  void  difference_of_gaussians(const Function& f,
				const float     center_wave_length,
				Function* const g,
				Support&        support){

    difference_of_gaussians(f,
			    center_wave_length*M_SQRT1_2*M_SQRT1_2,
			    center_wave_length*M_SQRT2*M_SQRT1_2,
			    g,
			    support);
  }


  
  template<typename Function>
  void difference_of_gaussians(const Function& f,
			       const float     small_wave_length,
			       const float     large_wave_length,
			       Function* const g,
			       Support&        support){

    // We need a temporary buffer.
    Support::buffer_type buffer = support.create_buffer();

    // Store f data in the fftw structures.
    support.load_space_data(f);
    
    // Compute the FFT.
    support.space_to_frequency();
    
    // Copy the result in a safe place.
    support.save_frequency_data_into_buffer(buffer);

    difference_of_gaussians(buffer,
			    small_wave_length,
			    large_wave_length,
			    g,
			    support);
  }


  
  template<typename Function>
  void difference_of_gaussians(const Function& f,
			       const float     small_wave_length,
			       const float     large_wave_length,
			       Function* const g){

    Support s(f.x_size(),f.y_size());

    difference_of_gaussians(f,small_wave_length,large_wave_length,g,s);
  }



  
  template<typename Function>
  void difference_of_gaussians(const Support::buffer_type& F,
			       const float                 center_wave_length,
			       Function* const             g,
			       Support&                    support){

    difference_of_gaussians(F,
			    center_wave_length*M_SQRT1_2,
			    center_wave_length*M_SQRT2,
			    g,
			    support);
  }

  
  

  
  
  template<typename Function>
  void difference_of_gaussians(const Support::buffer_type& F,
			       const float                 small_wave_length,
			       const float                 large_wave_length,
			       Function* const             g,
			       Support&                    support){

    using namespace Function_2D;


    // Type definition
    typedef float                              real_type;
    typedef std::complex<float>                complex_type;
    typedef FFT_normalized_gaussian            gaussian_type;
    typedef Minus<gaussian_type,gaussian_type> dog_type;

    // Creation of the data structures
    Array_2D<complex_type> dog_kernel(support.x_frequency_size(),
				      support.y_frequency_size(),
				      complex_type());

    const real_type width  = static_cast<real_type>(support.x_space_size());
    const real_type height = static_cast<real_type>(support.y_space_size());
    
    const real_type small_sigma_x = 2 * width / small_wave_length;
    const real_type small_sigma_y = 2 * height / small_wave_length;

    const real_type large_sigma_x = 2 * width / large_wave_length;
    const real_type large_sigma_y = 2 * height / large_wave_length;

    gaussian_type small_gaussian(0,0,small_sigma_x,small_sigma_y,0);
    gaussian_type large_gaussian(0,0,large_sigma_x,large_sigma_y,0);

    const dog_type dog(small_gaussian,large_gaussian);

    // Creation of DoG kernel
    FFT::fill_real_part(dog,&dog_kernel);

    // Load the data in the FFT engine
    support.load_frequency_data(dog_kernel);

    // Multiplication "in place" in the frequency space.    
    support.multiply_frequency_data_by_buffer(F,true);

    // Compute inverse FFT transform.
    support.frequency_to_space();

    // Return data into the original structures    
    support.save_space_data(g);
  }



  namespace Fourier_domain{ 

    void difference_of_gaussians(const Support::buffer_type& F,
				 const float                 center_wave_length,
				 Support::buffer_type* const G,
				 Support&                    support){

      FFT::Fourier_domain::difference_of_gaussians(F,
						   center_wave_length*M_SQRT1_2*M_SQRT1_2,
						   center_wave_length*M_SQRT2*M_SQRT1_2,
						   G,
						   support);
    }




    //! Duplication of the mix function to avoid back and forth copy from buffer.
    void difference_of_gaussians(const Support::buffer_type& F,
				 const float                 small_wave_length,
				 const float                 large_wave_length,
				 Support::buffer_type* const G,
				 Support&                    support){

      using namespace Function_2D;


      // Type definition
      typedef float                              real_type;
      typedef std::complex<float>                complex_type;
      typedef FFT_normalized_gaussian            gaussian_type;
      typedef Minus<gaussian_type,gaussian_type> dog_type;

      // Creation of the data structures
      Array_2D<complex_type> dog_kernel(support.x_frequency_size(),
					support.y_frequency_size(),
					complex_type());

      const real_type width  = static_cast<real_type>(support.x_space_size());
      const real_type height = static_cast<real_type>(support.y_space_size());
    
      const real_type small_sigma_x = 2 * width / small_wave_length;
      const real_type small_sigma_y = 2 * height / small_wave_length;

      const real_type large_sigma_x = 2 * width / large_wave_length;
      const real_type large_sigma_y = 2 * height / large_wave_length;

      gaussian_type small_gaussian(0,0,small_sigma_x,small_sigma_y,0);
      gaussian_type large_gaussian(0,0,large_sigma_x,large_sigma_y,0);

      const dog_type dog(small_gaussian,large_gaussian);

      // Creation of DoG kernel
      FFT::fill_real_part(dog,&dog_kernel);

      // Load the data in the FFT engine
      support.load_frequency_data(dog_kernel);

      // Multiplication "in place" in the frequency space.    
      support.multiply_frequency_data_by_buffer(F,false);

      // Return data into a buffer.
      support.save_frequency_data_into_buffer(*G);
    }

  }
  
}

#endif
